from unittest import TestCase

from musurgia.musurgia_exceptions import PdfAttributeError
from musurgia.musurgia_types import MusurgiaTypeError
from musurgia.pdf.drawobject import ClippingArea
from musurgia.pdf.labeled import TextLabel, Labeled
from musurgia.pdf.line import MarkLine, StraightLine
from musurgia.pdf.ruler import HorizontalRuler
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdfunit import PdfUnit
from musurgia.pdf.text import Text
from musurgia.tests.utils_for_tests import DummyMaster


class TestClippingArea(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.ruler = HorizontalRuler(length=2000)
        self.ca = ClippingArea(self.pdf, self.ruler)

    def test_get_row_height(self):
        x = self.ruler.get_height()
        assert self.ca.get_row_height() == x
        self.ruler.top_margin = 10
        assert self.ca.get_row_height() == x + 10

    def test_no_pdf(self):
        with self.assertRaises(PdfAttributeError):
            ClippingArea(pdf=None, draw_object=self.ruler).get_row_width()

        with self.assertRaises(PdfAttributeError):
            ClippingArea(pdf=None, draw_object=self.ruler).draw()


class TestPdfUnit(TestCase):
    def test_setting_global_unit(self):
        assert PdfUnit._DEFAULT_UNIT == "mm"
        assert PdfUnit.GLOBAL_UNIT == "mm"
        PdfUnit.GLOBAL_UNIT = "pt"
        assert PdfUnit.GLOBAL_UNIT == "pt"
        PdfUnit.reset()
        assert PdfUnit.GLOBAL_UNIT == "mm"
        with self.assertRaises(MusurgiaTypeError):
            PdfUnit.GLOBAL_UNIT = "bla"


class TestLabeled(TestCase):
    def test_no_master_error(self):
        t = TextLabel("bla")
        t.placement = "above"
        with self.assertRaises(AttributeError):
            assert t.placement == "above"

    def test_false_master_errr(self):
        t = TextLabel("bla", master=DummyMaster())
        l = Labeled()
        with self.assertRaises(AttributeError):
            l.add_text_label(t)
        t.master = l
        l.add_text_label(t)
        assert t.master == l

    def test_get_slave_position_wrong_type_or_slave_with_wrong_master(self):
        t = TextLabel("bla", master=DummyMaster())
        l = Labeled()
        with self.assertRaises(AttributeError):
            l.get_slave_position(t, "x")
        with self.assertRaises(TypeError):
            l.get_slave_position("slave", "y")
        t.master = l
        assert l.get_slave_position(t, "x") == 0


class TestSlaveErrors(TestCase):
    def test_no_master_error(self):
        ml = MarkLine(placement="start", mode="horizontal")
        with self.assertRaises(AttributeError):
            ml.relative_x

        with self.assertRaises(AttributeError):
            ml.relative_y

        with self.assertRaises(AttributeError):
            ml.top_margin

        with self.assertRaises(AttributeError):
            ml.left_margin

        with self.assertRaises(AttributeError):
            ml.bottom_margin

        with self.assertRaises(AttributeError):
            ml.right_margin

    def test_wrong_master(self):
        ml = MarkLine(placement="start", mode="horizontal")
        with self.assertRaises(TypeError):
            ml.master = TextLabel("something")


class TestNoPdfError(TestCase):
    def test_draw_with_no_pdf_error(self):
        t = Text("bla")
        with self.assertRaises(TypeError):
            t.draw(pdf=None)


class TestMarginsPositions(TestCase):
    def test_set_get_margins(self):
        l = StraightLine(mode="h", length=10)
        l.margins = (1, 2, 3, 4)
        assert l.get_margins() == {"bottom": 3.0, "left": 4.0, "right": 2.0, "top": 1.0}
        assert l.margins == (1, 2, 3, 4)

    def test_set_get_positions(self):
        l = StraightLine(mode="h", length=10)
        l.positions = (10, 20)
        assert l.get_positions() == {"x": 10.0, "y": 20.0}
        assert l.positions == (10, 20)



from abc import ABC
from typing import Any, Literal

from musurgia.musurgia_exceptions import RelativePositionNotSettableError
from musurgia.musurgia_types import (
    check_type,
    FontStyle,
    FontFamily,
    FontWeight,
    ConvertibleToFloat,
    VerticalPosition,
    HorizontalPosition,
    create_error_message,
)
from musurgia.pdf.drawobject import DrawObject
from musurgia.pdf.font import Font
from musurgia.pdf.margined import Margined
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdfunit import PdfUnit
from musurgia.pdf.positioned import Positioned

__all__ = ["Text", "PageText"]


class AbstractText(DrawObject, ABC):
    DEFAULT_FONT_FAMILY: FontFamily = "Courier"
    DEFAULT_FONT_SIZE: int = 10
    DEFAULT_FONT_WEIGHT: FontWeight = "medium"
    DEFAULT_FONT_STYLE: FontStyle = "regular"

    def __init__(
        self,
        value: Any,
        font_family: FontFamily = DEFAULT_FONT_FAMILY,
        font_weight: FontWeight = DEFAULT_FONT_WEIGHT,
        font_style: FontStyle = DEFAULT_FONT_STYLE,
        font_size: int = DEFAULT_FONT_SIZE,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.font = Font(font_family, font_weight, font_style, font_size)
        self._value: str
        self.value = value

    @property
    def font_family(self) -> FontFamily:
        return self.font.family

    @font_family.setter
    def font_family(self, val: FontFamily) -> None:
        # check_type(val, 'FontFamily', class_name=self.__class__.__name__, property_name='font_family')
        self.font.family = val

    @property
    def font_size(self) -> int:
        return self.font.size

    @font_size.setter
    def font_size(self, val: int) -> None:
        # check_type(val, int, class_name=self.__class__.__name__, property_name='font_size')
        self.font.size = val

    @property
    def font_weight(self) -> FontWeight:
        return self.font.weight

    @font_weight.setter
    def font_weight(self, val: FontWeight) -> None:
        # check_type(val, 'FontWeight', class_name=self.__class__.__name__, property_name='font_weight')
        self.font.weight = val

    @property
    def font_style(self) -> FontStyle:
        return self.font.style

    @font_style.setter
    def font_style(self, val: FontStyle) -> None:
        # check_type(val, 'FontStyle', class_name=self.__class__.__name__, property_name='font_style')
        self.font.style = val

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, val: Any) -> None:
        self._value = str(val)

    def get_text_width(self) -> float:
        return self.font.get_text_pixel_width(self.value) / PdfUnit.get_k()

    def get_text_height(self) -> float:
        return self.font.get_text_pixel_height(self.value) / PdfUnit.get_k()

    def get_relative_x2(self) -> float:
        return self.relative_x + self.get_text_width()

    def get_relative_y2(self) -> float:
        return self.relative_y + self.get_text_height()

    def draw(self, pdf: Pdf) -> None:
        if not isinstance(pdf, Pdf):
            raise TypeError(
                create_error_message(
                    message=f"{pdf} must be of type Pdf not {type(pdf)}",
                    class_name=self.__class__.__name__,
                    method_name="draw",
                    argument_name="pdf",
                )
            )
        if self.show:
            pdf.reset_font()
            style: Literal[
                "",
                "B",
                "I",
                "U",
                "BU",
                "UB",
                "BI",
                "IB",
                "IU",
                "UI",
                "BIU",
                "BUI",
                "IBU",
                "IUB",
                "UBI",
                "UIB",
            ]
            if self.font_style == "italic" and self.font_weight == "bold":
                style = "IB"
            elif self.font_style == "italic":
                style = "I"
            elif self.font_weight == "bold":
                style = "B"
            else:
                style = ""
            pdf.set_font(self.font.family, style=style, size=self.font_size)
            with pdf.pdf_draw_object_translate(self):
                pdf.text(x=0, y=0, text=self.value)


class Text(AbstractText, Positioned, Margined):
    pass


class PageText(Text):
    def __init__(
        self,
        value: Any,
        v_position: VerticalPosition = "top",
        h_position: HorizontalPosition = "left",
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(value=value, *args, **kwargs)  # type: ignore
        self._v_position: VerticalPosition
        self._h_position: HorizontalPosition
        self.v_position = v_position
        self.h_position = h_position

    @Text.relative_y.setter  # type: ignore
    def relative_y(self, val: ConvertibleToFloat) -> None:
        if val:
            raise RelativePositionNotSettableError

    @Text.relative_x.setter  # type: ignore
    def relative_x(self, val: ConvertibleToFloat) -> None:
        if val:
            raise RelativePositionNotSettableError

    @property
    def v_position(self) -> VerticalPosition:
        return self._v_position

    @v_position.setter
    def v_position(self, val: VerticalPosition) -> None:
        check_type(val, "VerticalPosition")
        self._v_position = val

    @property
    def h_position(self) -> HorizontalPosition:
        return self._h_position

    @h_position.setter
    def h_position(self, val: HorizontalPosition) -> None:
        check_type(val, "HorizontalPosition")
        self._h_position = val

    def draw(self, pdf: Pdf) -> None:
        pdf.reset_position()
        if self.h_position == "center":
            self._relative_x = (
                (pdf.w - pdf.l_margin - pdf.r_margin) / 2
            ) - self.get_width() / 2
        elif self.h_position == "left":
            self._relative_x = pdf.l_margin
        elif self.h_position == "right":
            self._relative_x = pdf.w - pdf.r_margin - self.get_width()
        else:
            raise NotImplementedError  # pragma: no cover

        if self.v_position == "top":
            self._relative_y = pdf.t_margin
        elif self.v_position == "bottom":
            self._relative_y = pdf.h - pdf.b_margin
        else:
            raise NotImplementedError  # pragma: no cover
        super().draw(pdf)



from fractions import Fraction
import math
from unittest import TestCase

from musicscore.util import xToD

from musurgia.utils import Normalizer, RelativeValueGenerator


class NormalizeTestCase(TestCase):
    def test_normalize_value_to_itself(self):
        normalize = Normalizer(normalization_range=(1, 5), input_value_range=(1, 5))
        for x in [1, 3, 5]:
            self.assertEqual(normalize.get_normalized_value(x), x)

    def test_normalize_value_to_itself_out_of_range(self):
        normalize = Normalizer(normalization_range=(1, 5), input_value_range=(1, 5))
        for x in [-2, 0, 1, 5, 7]:
            self.assertEqual(normalize.get_normalized_value(x), x)

    def test_normalize_value(self):
        normalize = Normalizer(normalization_range=(0, 1), input_value_range=(0, 10))
        for x in [0, 3, 5, 8, 10]:
            self.assertEqual(normalize.get_normalized_value(x), Fraction(x, 10))

    def test_normalize_value_out_of_range(self):
        normalize = Normalizer(normalization_range=(0, 1), input_value_range=(0, 10))
        for x in [-3, -2, 0, 10, 13, 15]:
            self.assertEqual(normalize.get_normalized_value(x), Fraction(x, 10))


class RelativeValueGeneratorTestCase(TestCase):
    def setUp(self):
        self.rvr = RelativeValueGenerator(
            value_range=(1, 10),
            value_grid=None,
            directions=[-1, -1, 1, 1],
            proportions=[2, 4, 3, 2],
        )

    def _get_directions(self, input_list):
        directions = []
        for index, x in enumerate(input_list[1:]):
            previous = input_list[index]
            if x > previous:
                directions.append(1)
            elif x < previous:
                directions.append(-1)
            else:
                directions.appedn(0)
        return directions

    def _get_proportions(self, input_list):
        intervals = xToD(input_list)
        factor = math.lcm(*[x.denominator for x in intervals])
        non_fraction_intervals = [factor * interval for interval in intervals]
        integer_intervals = [x.numerator for x in non_fraction_intervals]
        assert integer_intervals == non_fraction_intervals
        gcd = math.gcd(*integer_intervals)
        return [abs(x / gcd) for x in integer_intervals]

    def test_init(self):
        self.assertEqual(self.rvr.value_range, (1, 10))
        self.assertIsNone(self.rvr.value_grid)
        self.assertEqual(self.rvr.directions, [-1, -1, 1, 1])
        self.assertEqual(self.rvr.proportions, [2, 4, 3, 2])
        relative_values = self.rvr.get_values()
        self.assertListEqual(
            relative_values,
            [
                Fraction(10, 1),
                Fraction(7, 1),
                Fraction(1, 1),
                Fraction(11, 2),
                Fraction(17, 2),
            ],
        )
        self.assertListEqual(self._get_directions(relative_values), self.rvr.directions)
        self.assertEqual(min(relative_values), self.rvr.value_range[0])
        self.assertEqual(max(relative_values), self.rvr.value_range[1])
        self.assertListEqual(
            self._get_proportions(relative_values), self.rvr.proportions
        )



from unittest import TestCase
from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.tests.utils_for_tests import fractal_node_info
from musurgia.trees.timelinetree import TimelineDuration
from musurgia.utils import flatten


class TestFractalTreeReduceChildrenByCondition(TestCase):
    def test_reduce_first_layer(self):
        ft = FractalTimelineTree(
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        ft.add_layer()
        ft.add_layer()
        for node in ft.get_layer(1):
            node.reduce_children_by_condition(
                condition=lambda node: node.get_fractal_order() == 1
            )
        assert [node.get_fractal_order() for node in ft.iterate_leaves()] == [
            2,
            3,
            3,
            2,
            2,
            3,
        ]

    def test_value(self):
        ft = FractalTimelineTree(
            proportions=[1, 2, 3, 4, 5, 6],
            main_permutation_order=(4, 1, 5, 3, 6, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        ft.add_layer()
        ft.reduce_children_by_condition(
            lambda child: child.get_fractal_order() not in [2, 3]
        )
        self.assertEqual(
            [3, 2], [node.get_fractal_order() for node in ft.iterate_leaves()]
        )
        self.assertEqual([12, 8], [node.get_value() for node in ft.get_children()])


class TestFractalTreeReduceChildrenByNumberOfChildren(TestCase):
    def setUp(self):
        self.ft = FractalTimelineTree(
            proportions=(1, 2, 3, 4),
            main_permutation_order=(3, 1, 4, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        self.ft.add_layer()
        self.ft.add_layer()

        # print(self.ft.get_tree_representation(node_info))
        """
        └── 0:(1, 1): 20.0
            ├── 3:(2, 1): 6.0
            │   ├── 2:(3, 1): 1.2
            │   ├── 4:(3, 2): 2.4
            │   ├── 1:(3, 3): 0.6
            │   └── 3:(3, 4): 1.8
            ├── 1:(2, 2): 2.0
            │   ├── 3:(4, 1): 0.6
            │   ├── 1:(4, 2): 0.2
            │   ├── 4:(4, 3): 0.8
            │   └── 2:(4, 4): 0.4
            ├── 4:(2, 3): 8.0
            │   ├── 1:(1, 1): 0.8
            │   ├── 2:(1, 2): 1.6
            │   ├── 3:(1, 3): 2.4
            │   └── 4:(1, 4): 3.2
            └── 2:(2, 4): 4.0
                ├── 4:(2, 1): 1.6
                ├── 3:(2, 2): 1.2
                ├── 2:(2, 3): 0.8
                └── 1:(2, 4): 0.4
        """

    def test_reduce_backward(self):
        self.ft.reduce_children_by_size(size=3)
        for child in self.ft.get_children():
            child.reduce_children_by_size(size=3)
        # print(self.ft.get_tree_representation(node_info))
        assert (
            self.ft.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 3: (2, 1): 6.67
    │   ├── 2: (3, 1): 1.48
    │   ├── 4: (3, 2): 2.96
    │   └── 3: (3, 4): 2.22
    ├── 4: (2, 3): 8.89
    │   ├── 2: (1, 2): 1.98
    │   ├── 3: (1, 3): 2.96
    │   └── 4: (1, 4): 3.95
    └── 2: (2, 4): 4.44
        ├── 4: (2, 1): 1.98
        ├── 3: (2, 2): 1.48
        └── 2: (2, 3): 0.99
"""
        )

    def test_reduce_forwards(self):
        self.ft.reduce_children_by_size(size=3, mode="forwards")
        for child in self.ft.get_children():
            child.reduce_children_by_size(size=3, mode="forwards")
        # print(self.ft.get_tree_representation(node_info))
        assert (
            self.ft.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 3: (2, 1): 10.0
    │   ├── 2: (3, 1): 3.33
    │   ├── 1: (3, 3): 1.67
    │   └── 3: (3, 4): 5.0
    ├── 1: (2, 2): 3.33
    │   ├── 3: (4, 1): 1.67
    │   ├── 1: (4, 2): 0.56
    │   └── 2: (4, 4): 1.11
    └── 2: (2, 4): 6.67
        ├── 3: (2, 2): 3.33
        ├── 2: (2, 3): 2.22
        └── 1: (2, 4): 1.11
"""
        )

    def test_reduce_sieve(self):
        self.ft.reduce_children_by_size(size=3, mode="sieve")
        for child in self.ft.get_children():
            child.reduce_children_by_size(size=3, mode="sieve")

        # print(self.ft.get_tree_representation(node_info))
        assert (
            self.ft.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 1: (2, 2): 2.86
    │   ├── 1: (4, 2): 0.41
    │   ├── 4: (4, 3): 1.63
    │   └── 2: (4, 4): 0.82
    ├── 4: (2, 3): 11.43
    │   ├── 1: (1, 1): 1.63
    │   ├── 2: (1, 2): 3.27
    │   └── 4: (1, 4): 6.53
    └── 2: (2, 4): 5.71
        ├── 4: (2, 1): 3.27
        ├── 2: (2, 3): 1.63
        └── 1: (2, 4): 0.82
"""
        )

    def test_merge(self):
        ft_1 = FractalTimelineTree(
            proportions=(1, 2, 3, 4),
            main_permutation_order=(3, 1, 4, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        ft_2 = FractalTimelineTree(
            proportions=(1, 2, 3, 4),
            main_permutation_order=(3, 1, 4, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        ft_3 = FractalTimelineTree(
            proportions=(1, 2, 3, 4),
            main_permutation_order=(3, 1, 4, 2),
            duration=TimelineDuration(20),
            permutation_index=(1, 1),
        )
        for i in range(2):
            ft_1.add_layer()
            ft_2.add_layer()
            ft_3.add_layer()
        # print(self.ft.get_tree_representation(node_info))
        """
        └── 0: (1, 1): 20.0
            ├── 3: (2, 1): 6.0
            │   ├── 2: (3, 1): 1.2
            │   ├── 4: (3, 2): 2.4
            │   ├── 1: (3, 3): 0.6
            │   └── 3: (3, 4): 1.8
            ├── 1: (2, 2): 2.0
            │   ├── 3: (4, 1): 0.6
            │   ├── 1: (4, 2): 0.2
            │   ├── 4: (4, 3): 0.8
            │   └── 2: (4, 4): 0.4
            ├── 4: (2, 3): 8.0
            │   ├── 1: (1, 1): 0.8
            │   ├── 2:(1, 2): 1.6
            │   ├── 3: (1, 3): 2.4
            │   └── 4: (1, 4): 3.2
            └── 2: (2, 4): 4.0
                ├── 4: (2, 1): 1.6
                ├── 3: (2, 2): 1.2
                ├── 2: (2, 3): 0.8
                └── 1: (2, 4): 0.4
        """
        assert ft_1._get_merge_lengths(size=2, merge_index=0) == [3, 1]
        assert ft_2._get_merge_lengths(size=2, merge_index=2) == [2, 2]
        assert ft_3._get_merge_lengths(size=2, merge_index=3) == [3, 1]

        for index, ft in zip([0, 2, 3], [ft_1, ft_2, ft_3]):
            ft.reduce_children_by_size(size=2, mode="merge", merge_index=index)
            for child in ft.get_children():
                child.reduce_children_by_size(size=3, mode="merge", merge_index=index)
        # print(ft_1.get_tree_representation(node_info))
        assert (
            ft_1.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 3: (2, 1): 16.0
    │   ├── 2: (3, 1): 9.6
    │   ├── 1: (3, 3): 1.6
    │   └── 3: (3, 4): 4.8
    └── 2: (2, 4): 4.0
        ├── 4: (2, 1): 2.8
        ├── 2: (2, 3): 0.8
        └── 1: (2, 4): 0.4
"""
        )
        # print(ft_2.get_tree_representation(node_info))
        assert (
            ft_2.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 3: (2, 1): 8.0
    │   ├── 2: (3, 1): 1.6
    │   ├── 4: (3, 2): 3.2
    │   └── 1: (3, 3): 3.2
    └── 4: (2, 3): 12.0
        ├── 1: (1, 1): 1.2
        ├── 2: (1, 2): 2.4
        └── 3: (1, 3): 8.4
"""
        )
        # print(ft_3.get_tree_representation(node_info))
        assert (
            ft_3.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 20.0
    ├── 3: (2, 1): 16.0
    │   ├── 2: (3, 1): 9.6
    │   ├── 1: (3, 3): 1.6
    │   └── 3: (3, 4): 4.8
    └── 2: (2, 4): 4.0
        ├── 4: (2, 1): 2.8
        ├── 2: (2, 3): 0.8
        └── 1: (2, 4): 0.4
"""
        )

        for ft in [ft_1, ft_2, ft_3]:
            assert (
                ft.get_value()
                == sum(flatten(ft.get_layer(1, key=lambda node: node.get_value())))
                == sum(flatten(ft.get_layer(2, key=lambda node: node.get_value())))
            )
